"""  
Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors

SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import requests
import logging
import json
import datetime
from base64 import b64encode


class RestClient():

    def __init__(self, agent):
        self.logger = logging.getLogger(__name__)
        self.serial = agent.serial
        self.configuration = agent.configuration
        self.base_url = agent.url
        if not self.base_url.startswith('http'):
            self.base_url = f'https://{self.base_url}'
        self.token = agent.token

    def get_auth_header(self):
        if self.token:
            return {'Authorization': 'Bearer '+self.token}
        else:
            credentials = self.configuration.getCredentials()
            self.tenant = credentials[0]
            self.user = credentials[1]
            self.password = credentials[2]
            auth_string = f'{self.tenant}/{self.user}:{self.password}'
            encoded_auth_string = b64encode(
                bytes(auth_string, 'utf-8')).decode('ascii')
            return {'Authorization': 'Basic '+encoded_auth_string, }

    def update_managed_object(self, internal_id, payload):
        #self.logger.info('Update of managed Object')
        try:
            url = f'{self.base_url}/inventory/managedObjects/{internal_id}'
            headers = self.get_auth_header()
            headers['Content-Type'] = 'application/json'
            self.logger.debug(f'Sending Request to url {url}')
            response = requests.request(
                "PUT", url, headers=headers, data=payload)
            self.logger.debug('Response from request: ' + str(response.text))
            self.logger.debug(
                'Response from request with code : ' + str(response.status_code))
            if response.status_code == 200 or response.status_code == 201:
                #self.logger.info('Managed object updated in C8Y')
                return True
            else:
                self.logger.warning('Managed object not updated in C8Y')
                return False
        except Exception as e:
            self.logger.error('The following error occured: %s' % (str(e)))

    def get_internal_id(self, external_id):
        try:
            #self.logger.info('Checking against indentity service what is internalID in C8Y')
            url = f'{self.base_url}/identity/externalIds/c8y_Serial/{external_id}'
            self.logger.debug(f'Sending Request to url {url}')
            headers = self.get_auth_header()
            headers['Content-Type'] = 'application/json'
            headers['Accept'] = 'application/json'
            response = requests.request("GET", url, headers=headers)
            self.logger.debug('Response from request: ' + str(response.text))
            self.logger.debug(
                'Response from request with code : ' + str(response.status_code))
            if response.status_code == 200:
                #self.logger.info('Managed object exists in C8Y')
                json_data = json.loads(response.text)
                internalID = json_data['managedObject']['id']
                #self.logger.info("The internalID for " + str(external_id) + " is " + str(internalID))
                self.logger.debug('Returning the internalID')
                return internalID
            else:
                self.logger.warning(
                    'Response from request: ' + str(response.text))
                self.logger.warning('Got response with status_code: ' +
                                    str(response.status_code))
                return None
        except Exception as e:
            self.logger.error('The following error occured: %s' % (str(e)))
            return None

    def upload_binary_logfile(self, internal_id, payload, file):
        #self.logger.info('Update of managed Object')
        try:
            url = f'{self.base_url}/inventory/binaries'
            headers = self.get_auth_header()
            headers['Content-Type'] = 'multipart/form-data'
            headers['Accept'] = 'application/json'
            self.logger.debug(f'Sending Request to url {url}')
            response = requests.request(
                "POST", url, headers=headers, data=payload, files=file)
            print("Responsestatuscode:" + str(response.status_code))
            print("RESPONSEMSG: "+str(response.text))
            self.logger.debug(
                'Response from request: ' + str(response.text))
            self.logger.debug(
                'Response from request with code : ' + str(response.status_code))
            if response.status_code == 200 or response.status_code == 201:
                json_data = json.loads(response.text)
                binaryurl = json_data["self"]
                # print(binaryurl)
                # return binaryurl
                return binaryurl
            else:
                self.logger.warning('Binary upload failed in C8Y')
                return False
        except Exception as e:
            self.logger.error('The following error occured: %s' % (str(e)))

            
    def create_logfile_event(self, mo_id):
        try:
            url = f'{self.base_url}/event/events'
            headers = self.get_auth_header()
            headers['Content-Type'] = 'application/json'
            headers['Accept'] = 'application/json'
            payload = {
                "time" : datetime.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"),
                "type" : "c8y_LogfileRequest",
                "text" : "LogFile Request Event",
                "source": { "id" : mo_id}
            }
            self.logger.debug(f'Sending Request to url {url}')
            response = requests.request(
                "POST", url, headers=headers, data=json.dumps(payload))
            self.logger.debug(
                'Response from request: ' + str(response.text))
            self.logger.debug(
                'Response from request with code : ' + str(response.status_code))
            if response.status_code == 200 or response.status_code == 201:
                json_data = json.loads(response.text)
                event_id = json_data["id"]
                # print(binaryurl)
                # return binaryurl
                return event_id
            else:
                self.logger.warning('Creating LogFileEvent failed!')
                return None
        except Exception as ex:
            self.logger.error('The following error occured: %s' % (str(ex)))
            return None
        
    def create_configfile_event(self, mo_id, configtype, path):
        try:
            url = f'{self.base_url}/event/events'
            headers = self.get_auth_header()
            headers['Content-Type'] = 'application/json'
            headers['Accept'] = 'application/json'
            payload = {
                "time" : datetime.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"),
                "type" : configtype,
                "text" : path,
                "description": "Config File Snapshot Request Event",
                "source": { "id" : mo_id}
            }
            self.logger.debug(f'Sending Request to url {url}')
            response = requests.request(
                "POST", url, headers=headers, data=json.dumps(payload))
            self.logger.debug(
                'Response from request: ' + str(response.text))
            self.logger.debug(
                'Response from request with code : ' + str(response.status_code))
            if response.status_code == 200 or response.status_code == 201:
                json_data = json.loads(response.text)
                event_id = json_data["id"]
                # print(binaryurl)
                # return binaryurl
                return event_id
            else:
                self.logger.warning('Creating LogFileEvent failed!')
                return None
        except Exception as ex:
            self.logger.error('The following error occured: %s' % (str(ex)))
            return None

    def upload_event_logfile(self, mo_id, payload, file):
        #self.logger.info('Update of managed Object')
        try:
            event_id = self.create_logfile_event(mo_id)
            if not event_id:
                return None
            url = f'{self.base_url}/event/events/{event_id}/binaries'
            headers = self.get_auth_header()
            headers['Content-Type'] = 'multipart/form-data'
            headers['Accept'] = 'application/json'
            self.logger.debug(f'Sending Request to url {url}')
            response = requests.request(
                "POST", url, headers=headers, data=payload, files=file)
            self.logger.debug('Response from request: ' + str(response.text))
            self.logger.debug(
                'Response from request with code : ' + str(response.status_code))
            if response.status_code == 200 or response.status_code == 201:
                json_data = json.loads(response.text)
                binaryurl = json_data["self"]
                # print(binaryurl)
                # return binaryurl
                return binaryurl
            else:
                self.logger.warning('Binary upload failed in C8Y')
                return None
        except Exception as e:
            self.logger.error('The following error occured: %s' % (str(e)))
    
    def upload_event_configfile(self, mo_id, payload, file, configtype, path):
        #self.logger.info('Update of managed Object')
        try:
            event_id = self.create_configfile_event(mo_id,configtype,path)
            if not event_id:
                return None
            url = f'{self.base_url}/event/events/{event_id}/binaries'
            headers = self.get_auth_header()
            headers['Content-Type'] = 'multipart/form-data'
            headers['Accept'] = 'application/json'
            self.logger.debug(f'Sending Request to url {url}')
            response = requests.request(
                "POST", url, headers=headers, data=payload, files=file)
            self.logger.debug('Response from request: ' + str(response.text))
            self.logger.debug(
                'Response from request with code : ' + str(response.status_code))
            if response.status_code == 200 or response.status_code == 201:
                json_data = json.loads(response.text)
                binaryurl = json_data["self"]
                # print(binaryurl)
                # return binaryurl
                return binaryurl
            else:
                self.logger.warning('Binary upload failed in C8Y')
                return None
        except Exception as e:
            self.logger.error('The following error occured: %s' % (str(e)))
        
    def download_c8y_binary(self, url, file):
        #self.logger.info('Update of managed Object')
        try:
            headers = self.get_auth_header()
            headers['Content-Type'] = 'multipart/form-data'
            headers['Accept'] = 'application/json'
            self.logger.info(f'Sending Request to url {url}')
            response = requests.request(
                "GET", url, headers=headers, allow_redirects=True)
            self.logger.debug('Response from request: ' + str(response.text))
            self.logger.debug('Response from request with code : ' + str(response.status_code))
            if response.status_code == 200 or response.status_code == 201:
                open(file,'wb').write(response.content)
                # print(binaryurl)
                # return binaryurl
                return file
            else:
                self.logger.warning('Binary download failed in C8Y')
                return None
        except Exception as e:
            self.logger.error('The following error occured: %s' % (str(e)))
            return None

    def get_all_dangling_operations(self, internal_id):
        try:
            url = f'{self.base_url}/devicecontrol/operations?status=EXECUTING&deviceId={internal_id}'
            headers = self.get_auth_header()
            headers['Content-Type'] = 'application/json'
            headers['Accept'] = 'application/json'
            response = requests.request("GET", url, headers=headers)
            self.logger.debug('Response from request: ' + str(response.text))
            self.logger.debug(
                'Response from request with code : ' + str(response.status_code))
            if response.status_code == 200:
                json_data = json.loads(response.text)
                #operations = []
                operations = json_data['operations']
                #self.logger.info("The internalID for " + str(external_id) + " is " + str(internalID))
                self.logger.debug(
                    f'Returning Operations with status EXECUTING {operations}')
                return operations
            else:
                self.logger.warning(
                    'Response from request: ' + str(response.text))
                self.logger.warning('Got response with status_code: ' +
                                    str(response.status_code))
                return None
        except Exception as e:
            self.logger.error('The following error occured: %s' % (str(e)))

    def set_operations_to_failed(self, operations):
        if len(operations) > 0:
            try:
                for op in operations:
                    url = f'{self.base_url}/devicecontrol/operations/{op["id"]}'
                    headers = self.get_auth_header()
                    headers['Content-Type'] = 'application/json'
                    headers['Accept'] = 'application/json'
                    payload = {
                        "status": "FAILED",
                        "failureReason": "Operation unexpectedly interrupted. Check logs for details"
                    }
                    response = requests.request(
                        "PUT", url, headers=headers, data=json.dumps(payload))
                    self.logger.debug(
                        'Response from request: ' + str(response.text))
                    self.logger.debug(
                        'Response from request with code : ' + str(response.status_code))
                    if response.status_code == 200:
                        return True
                    else:
                        return False
            except Exception as e:
                self.logger.error('The following error occured: %s' % (str(e)))

    def create_SmartRest_template(self,template,template_id):
        try:
            url = f'{self.base_url}/inventory/managedObjects'
            self.logger.debug(f'Sending Request to url {url}')
            payload = json.loads(template)
            headers = self.get_auth_header()
            headers['Content-Type'] ='application/json'
            headers['Accept'] = 'application/json'
            response = requests.request("POST", url, headers=headers, data = json.dumps(payload))
            if response.status_code == 200 or response.status_code==201:
                json_data = json.loads(response.text)
                self.logger.info(f'Template created with id {json_data["id"]}')
                payload = json.loads(f'{{"externalId": "{template_id}","type": "c8y_SmartRest2DeviceIdentifier"}}')
                url = f'{self.base_url}/identity/globalIds/{json_data["id"]}/externalIds'
                self.logger.debug(f'Sending Request for idenenity of smart rest template to url {url}')
                response = requests.request("POST", url, headers=headers, data = json.dumps(payload))
                if response.status_code == 200 or response.status_code==201:
                    self.logger.debug('Response from request of identity API for smart rest template: ' + str(response.text))
                    return True
                else:
                    self.logger.warning('Response from request: ' + str(response.text))
                    self.logger.warning('Got response with status_code: ' +
                            str(response.status_code))
                    return False
            else:
                self.logger.warning('Response from request: ' + str(response.text))
                self.logger.warning('Got response with status_code: ' +
                            str(response.status_code))
                return False
        except Exception as e:
            self.logger.error('The following error occured while trying to create SmartRest template: %s' % (str(e)))    

    def check_SmartRest_template_exists(self,templateID):
        try:
            url = f'{self.base_url}/identity/externalIds/c8y_SmartRest2DeviceIdentifier/{templateID}'
            self.logger.debug(f'Sending Request to url {url}')
            headers = self.get_auth_header()
            headers['Content-Type'] ='application/json'
            headers['Accept'] = 'application/json'
            response = requests.request("GET", url, headers=headers)
            self.logger.info('Checking against indentity service')
            if response.status_code == 200:
                self.logger.info('Managed object exists in C8Y')
                self.logger.debug('Returning the internalID')
                json_data = json.loads(response.text)
                return True
            else:
                self.logger.warning('Response from request: ' + str(response.text))
                self.logger.warning('Got response with status_code: ' + str(response.status_code))
                return False
        except Exception as e:
            self.logger.error('The following error occured while trying to check for existing SmartRest templates: %s' % (str(e)))
            return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""  
Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors

SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import logging, io, re
from datetime import datetime
from c8ydm.framework.modulebase import Initializer, Listener
from c8ydm.framework.smartrest import SmartRESTMessage

class LogfileInitializer(Initializer, Listener):
    logger = logging.getLogger(__name__)
    fragment = 'c8y_LogfileRequest'
    

    def getMessages(self):
        msg = SmartRESTMessage('s/us', '118', ['agentlog'])
        return [msg]

    def getSupportedOperations(self):
        return ['c8y_LogfileRequest']
    
    def getSupportedTemplates(self):
        return []

    def _set_executing(self):
        executing = SmartRESTMessage('s/us', '501', [self.fragment])
        print("Executing MSG send")
        self.agent.publishMessage(executing)

    #datei anhängen
    def _set_success(self, url):
        success = SmartRESTMessage('s/us', '503', [self.fragment, url])
        print("Success MSG send")
        self.agent.publishMessage(success)

    def _set_failed(self, reason):
        failed = SmartRESTMessage('s/us', '502', [self.fragment, reason])
        self.agent.publishMessage(failed)
    
    def handleOperation(self, message):
        mo_id = self.agent.rest_client.get_internal_id(self.agent.serial)  
        try:
            if 's/ds' in message.topic and message.messageId == '522':
                # When multiple operations received just take the first one for further processing
                #self.logger.debug("message received :" + str(message.values))
                deviceid = message.values[0]
                #logtype = message.values[1]
                starttime = message.values[2]
                endtime = message.values[3]
                searchtext = message.values[4]
                searchtext = searchtext.lower()
                maximumlines = message.values[5]
                #print('\n deviceid: {}\n starttime: {}\n endtime: {}\n searchtext:{}\n maximumlines:{} \n'.format(deviceid, starttime, endtime, searchtext, maximumlines))
                self._set_executing()
                #self.logger.info('LogFile HandleOperation Called for '+ deviceid)
                starttime = starttime.replace("T", " ")
                endtime = endtime.replace("T", " ")
                starttime = datetime.fromisoformat(starttime[:16])
                endtime = datetime.fromisoformat(endtime[:16])
                fileLines = []
                searchtextindata = False
                path = self.agent.path
                with open(path / 'agent.log', 'r') as f:
                    for line in f:
                        stripped_line = line.strip().lower()
                        fileLines.append(stripped_line)

                        if(searchtext in line and searchtext !=''):
                            searchtextindata = True

                    if searchtextindata == True:
                        num_lines = len(fileLines)
                        newOutput = ''
                        outputfound = False
                        while(outputfound == False):
                            for index, line in enumerate(fileLines):
                                linematch = re.match("[0-9][0-9][0-9][0-9][-][0-9][0-9]+", line[:16])
                                linehaslogtime = bool(linematch)
                                if(linehaslogtime == True and outputfound == False):
                                    logtime = datetime.fromisoformat(line[:16])
                                    # #self.logger.debug('Starttime: {}\nLogtime: {}\nEndtime:{}'.format(starttime,logtime,endtime))
                                    if starttime <logtime <endtime:
                                        if searchtext in line:
                                            i = 0
                                            while(i<int(maximumlines) and index+i < num_lines):
                                                newOutput+= fileLines[index+i] + '\n'
                                                i+=1
                                                outputfound = True
                        memFile = io.BytesIO(newOutput.encode('utf8'))
                        payload = {'object' : '{"name" : "logfile'+ deviceid+'", "type" : "text/plain" }'}
                        file = [('file' , memFile)]
                        binaryurl = self.agent.rest_client.upload_event_logfile(mo_id, payload, file)
                        if binaryurl:
                            self._set_success(binaryurl)
                            self.logger.debug("LogHandler uploaded Binary under following URL: "+binaryurl)
                        else:
                            self._set_failed('Could not upload logfile')
                                            
                    elif searchtext=='':
                        num_lines = len(fileLines)
                        newOutput = ''
                        outputfound = False
                        while(outputfound == False):
                            for index, line in enumerate(fileLines):
                                linematch = re.match("[0-9][0-9][0-9][0-9][-][0-9][0-9]+", line[:16])
                                linehaslogtime = bool(linematch)
                                if(linehaslogtime == True and outputfound == False):
                                    logtime = datetime.fromisoformat(line[:16])
                                    #self.logger.debug('Starttime: {}\nLogtime: {}\nEndtime:{}'.format(starttime,logtime,endtime))
                                    if starttime <logtime <endtime:
                                        i = 0
                                        while(i<int(maximumlines) and index+i < num_lines):
                                            newOutput+= fileLines[index+i] + '\n'
                                            i+=1
                                            outputfound = True
                        memFile = io.BytesIO(newOutput.encode('utf8'))
                        payload = {'object' : '{"name" : "logfile'+ deviceid+'", "type" : "text/plain" }'}
                        file = [('file' , memFile)]
                        binaryurl = self.agent.rest_client.upload_event_logfile(mo_id, payload, file)
                        if binaryurl:
                            self._set_success(binaryurl)
                            self.logger.debug("LogHandler uploaded Binary under following URL: "+binaryurl)
                        else:
                            self._set_failed('Could not upload logfile')
                    else:
                        #print('Searchstring is not inside file.')
                        self._set_failed('Searchstring is not inside file')
                #after the 'with' statement everything is closed to free the memorybuffer
                f.close()
                self.logger.debug("logfilerequest handled")
        except Exception as e:
            self._set_failed(e)




"""  
Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors

SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
"""Cloud Remote Access Listener"""
import logging
from typing import List, Optional

from requests.auth import HTTPBasicAuth
from c8ydm.framework.modulebase import Listener
from c8ydm.framework.smartrest import SmartRESTMessage
from c8ydp.device_proxy import DeviceProxy, WebSocketFailureException
from c8ydm.utils import Configuration

class RemoteAccessListener(Listener):
    """Establishes Remote Access Connections"""
    xid = 'c8y-dm-agent-v1.0'
    logger = logging.getLogger(__name__)
    remote_access_op_template = 'dm600'
    remote_access_default_template = '530'
    fragment = 'c8y_RemoteAccessConnect'

    def _set_executing(self):
        executing = SmartRESTMessage('s/us', '501', [self.fragment])
        self.agent.publishMessage(executing)

    def _set_success(self):
        success = SmartRESTMessage('s/us', '503', [self.fragment])
        self.agent.publishMessage(success)

    def _set_failed(self, reason):
        failed = SmartRESTMessage('s/us', '502', [self.fragment, reason])
        self.agent.publishMessage(failed)

    def handleOperation(self, message):
        """Callback that is executed for any operation received

            Raises:
            Exception: Error when handling the operation
       """
        try:
            #self.logger.debug(
            #    f'Handling Cloud Remote Access operation: listener={__name__}, message={message}')

            if message.messageId == self.remote_access_default_template or message.messageId == self.remote_access_op_template:
                self._set_executing()
                self._proxy_connect(message)
            return

        except Exception as ex:
            self.logger.error(f'Handling operation error. exception={ex}')
            self._set_failed(str(ex))
            # raise

    def _proxy_connect(self, message):
        """
        Creates the Device Proxy and connects to WebSocket and TCP Port
        """
        tcp_host = message.values[1]
        tcp_port = int(message.values[2])
        connection_key = message.values[3]

        config = self.agent.configuration
        credentials = config.getCredentials()

        token = self.agent.token
        tenantuser = credentials[0]+'/'+ credentials[1]
        password = credentials[2]

        if token is None and tenantuser is None and password is None:
            raise WebSocketFailureException(
                'OAuth Token or tenantuser and password must be provided!')
        base_url = config.getValue('mqtt', 'url')
         # Not sure which buffer size is good, starting with 16 KB (16 x 1024)
        #buffer_size = self.utils.config.getint('remote-connect', 'tcp.buffer.size')
        self._device_proxy = DeviceProxy(
            tcp_host, tcp_port, None, connection_key, base_url, tenantuser, password, token, None)
        self._device_proxy.connect()
        self._set_success()

    def getSupportedOperations(self):
        return [self.fragment]

    def getSupportedTemplates(self):
        return [self.xid]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""  
Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors

SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import os
import signal
import subprocess
import sys
import time
import pathlib
from os.path import expanduser
from logging.handlers import RotatingFileHandler
import signal

from paho.mqtt.client import LOGGING_LEVEL
import c8ydm.utils.systemutils as systemutils
from c8ydm.client import Agent
from c8ydm.client import Bootstrap
from c8ydm.utils import Configuration

agent = None
bootstrap_agent = None
terminated = False
simulated = False

def handle_sigterm(*args):
    global terminated
    if not terminated:
        raise KeyboardInterrupt

def keyboard_interupt_hook(exctype, value, traceback):
    try:
        global terminated
        if exctype == KeyboardInterrupt:
            if not terminated:
                logging.info(f'KeyboardInterrupt called!')
                terminated = True
                global agent
                if agent:
                    agent.stop()
                global bootstrap_agent
                if bootstrap_agent:
                    bootstrap_agent.stop()
                stop()
                sys.exit(0)
        #else:
        #    sys.__excepthook__(exctype, value, traceback)
    except Exception as ex:
        logging.error(ex)


def start():
    try:
        sys.excepthook = keyboard_interupt_hook
        global agent
        global simulated
        agent = None
        signal.signal(signal.SIGTERM, handle_sigterm)
        home = expanduser('~')
        logger = logging.getLogger()
        logger.setLevel(logging.INFO)
        path = pathlib.Path(home + '/.cumulocity')
        path.mkdir(parents=True, exist_ok=True)
        config_path = pathlib.Path(path / 'agent.ini')
        if not config_path.is_file():
            sys.exit(f'No agent.ini found in "{path}". Create it to properly configure the agent.')
        config = Configuration(str(path))
        loglevel = config.getValue('agent', 'loglevel')
        logger.setLevel(loglevel)
        log_file_formatter = logging.Formatter(
            '%(asctime)s %(threadName)s %(levelname)s %(name)s %(message)s')
        log_console_formatter = logging.Formatter('%(asctime)s %(threadName)s %(levelname)s %(name)s %(message)s')
        # Set default log format
        if len(logger.handlers) == 0:
            console_handler = logging.StreamHandler()
            console_handler.setFormatter(log_console_formatter)
            console_handler.setLevel(loglevel)
            logger.addHandler(console_handler)
        else:
            handler = logger.handlers[0]
            handler.setFormatter(log_console_formatter)

        # Max 5 log files each 10 MB.
        rotate_handler = RotatingFileHandler(filename=path / 'agent.log', maxBytes=10000000,
                                            backupCount=5)
        rotate_handler.setFormatter(log_file_formatter)
        rotate_handler.setLevel(loglevel)
        # Log to Rotating File
        logger.addHandler(rotate_handler)

        containerId = None
        serial = None

        if serial == None:
            logging.debug(f'Serial not proviced. Fetching from system...')
            try:
                if os.getenv('CONTAINER') == 'docker':
                    containerId = subprocess.check_output(['bash', '-c', 'hostname'], universal_newlines=True)
                    containerId = containerId.strip('\n')
                    containerId = containerId.strip()
                    logging.info('Container Id: %s', str(containerId))
                    simulated = True
            except Exception as e:
                logging.error('Could not retrieve container Id: ' + str(e))

            if containerId is None:
                serial = systemutils.getSerial()
                logging.info('Output of SystemUtils Serial: %s', serial)
                simulated = False
            else:
                serial = containerId.strip()
                simulated = True
        if config.getValue('agent','device.id'):
            serial = config.getValue('agent','device.id')
        #if not simulated:
        startDaemon(str(path) + '/agent.pid')
        logging.info(f'Serial: {serial}')

        credentials = config.getCredentials()
        logging.debug('Credentials:')
        logging.debug(credentials)
        agent = Agent(serial, path, config, str(path) + '/agent.pid', simulated)
        cert_auth = config.getBooleanValue('mqtt','cert_auth')
        logging.debug(f'cert_auth: {cert_auth}')
        if not cert_auth and credentials is None:
            logging.info('No credentials found. Starting bootstrap mode.')
            bootstrapCredentials = config.getBootstrapCredentials()
            if bootstrapCredentials is None:
                logging.error('No bootstrap credentials found. Stopping agent.')
                return
            bootstrapAgent = Bootstrap(serial, str(path), config)
            bootstrapAgent.bootstrap()
            credentials = config.getCredentials()
            if credentials is None:
                logging.error('No credentials found after bootstrapping. Stopping agent.')
                return        
        agent.run()
    except Exception as ex:
        logger.exception(f'Error on main start {ex}', ex)
        


def stop():
    path = expanduser('~') + '/.cumulocity'
    stopDaemon(path + '/agent.pid')

def stopDaemon(pidfile):
    """Stop the daemon."""
    logging.info(f'Stopping...')
    global terminated
    # Get the pid from the pidfile
    try:
        with open(pidfile, 'r') as pf:
            pid = int(pf.read().strip())
    except IOError:
        pid = None

    if not pid:
        return
    else:
        delpid(pidfile)
    # Try killing the daemon process
    try:
        while 1:
            #if not terminated:
            logging.debug(f'Try killing pid {pid}')
            os.kill(pid, signal.SIGKILL)
            time.sleep(0.1)
                
    except OSError as err:
        e = str(err.args)
        if e.find("No such process") > 0:
            if os.path.exists(pidfile):
                os.remove(pidfile)
        else:
            print(str(err.args))
            sys.exit(1)
    


def delpid(pidfile):
    if pathlib.Path(pidfile).is_file():
        logging.info(f'Removing PID File {pidfile}')
        os.remove(pidfile)


def startDaemon(pidfile):
    """Start the daemon."""
    print("Starting...")
    # Check for a pidfile to see if the daemon already runs
    pid = None
    global simulated
    try:
        with open(pidfile, 'r') as pf:
            pid = int(pf.read().strip())
            logging.info(f'Found pid file with pid {str(pid)}')
    except IOError:
        pid = None

    if pid:
        if simulated or not isPidRunning(pid):
            delpid(pidfile)
            pid = str(os.getpid())
            with open(pidfile, 'w+') as f:
                f.write(pid + '\n')
        else:
            message = "pidfile {0} already exist. " + \
                      "Daemon already running? Stop it with sudo c8ydm.stop first.\n"
            sys.stderr.write(message.format(pidfile))
            sys.exit(1)
    else:
        pid = str(os.getpid())
        with open(pidfile, 'w+') as f:
            f.write(pid + '\n')


def isPidRunning(pid):
    """ Check For the existence of a unix pid. """
    try:
        logging.info('Checking if pid ' + str(pid) + ' is existing...')
        os.kill(pid, 0)
        logging.info('Pid ' + str(pid) + ' exists')
    except OSError:
        logging.info('Pid ' + str(pid) + ' does not exist')
        return False
    else:
        return True

if __name__ == '__main__':
    start()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""  
Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors

SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import configparser, logging, os
from shutil import copyfile
from configparser import NoOptionError, NoSectionError

class Configuration():
  logger = logging.getLogger(__name__)
  credentialsCategory = 'secret'
  bootstrapTenant = 'c8y.bootstrap.tenant'
  bootstrapUser = 'c8y.bootstrap.user'
  bootstrapPassword = 'c8y.bootstrap.password'
  tenant = 'c8y.tenant'
  user = 'c8y.username'
  password = 'c8y.password'

  def __init__(self, path):
    self.configPath = path + '/agent.ini'
    self.configuration = configparser.ConfigParser()
    self.readFromFile()
    self.overrideFromEnv()

  def readFromFile(self):
    self.configuration.read(self.configPath)

  def overrideFromEnv(self):
    """
    expands environment variables starting with `C8Y` in values.
    Mapping rules from environment variable name to config key:

    - Prefix C8YDM_<PREFIX>_ means what section the option belongs to
    - Upper case letters are mapped to lower case letters
    - Double underscore __ is mapped to .

    e.g. `C8YDM_SECRET_C8Y__TENANT` is mapped to `c8y.tenant` in [secret] section.
    """
    prefix = 'C8YDM_'
    envs = {k.replace(prefix, '', 1): v for k, v in os.environ.items() if k.startswith(prefix)}
    for key, value in envs.items():
      # note: '_' in category is not allowed
      category, key0 = key.lower().split('_', 1)
      key1 = key0.replace('__', '.')
      # todo: print override env in debug log. you need to modify main.py
      # because log level is known only after the creation of this object
      self.setValue(category, key1, value)

  def getValue(self, category, key):
    try:
      return self.configuration.get(category, key)
    except (NoOptionError, NoSectionError):
      return None
  
  def getBooleanValue(self, category, key):
    try:
      return self.configuration.getboolean(category, key)
    except (NoOptionError, NoSectionError):
      return None

  def setValue(self, category, key, value):
    if category not in self.configuration.sections():
      self.configuration.add_section(category)
    self.configuration.set(category, key, value)
    with open(self.configPath, 'w') as cfgfile:
      self.configuration.write(cfgfile)

  def getBootstrapCredentials(self):
    tenant = self.getValue(self.credentialsCategory, self.bootstrapTenant)
    user = self.getValue(self.credentialsCategory, self.bootstrapUser)
    password = self.getValue(self.credentialsCategory, self.bootstrapPassword)

    if tenant is not None and user is not None and password is not None:
      return [tenant, user, password]
    return None

  def getCredentials(self):
    tenant = self.getValue(self.credentialsCategory, self.tenant)
    user = self.getValue(self.credentialsCategory, self.user)
    # if the password contains % it needs to be escaped as % has special meaning in configparser
    password = self.getValue(self.credentialsCategory, self.password.replace('%','%%'))

    if tenant is not None and user is not None and password is not None:
      return [tenant, user, password]
    return None

  def writeCredentials(self, tenant, user, password):
    #TODO Write this in a secure storage.
    self.configuration.set(self.credentialsCategory, self.tenant, tenant)
    self.configuration.set(self.credentialsCategory, self.user, user)
    self.configuration.set(self.credentialsCategory, self.password, password.replace('%', '%%'))
    with open(self.configPath, 'w') as cfgfile:
      self.configuration.write(cfgfile)

  def getConfigString(self):
    configs = []
    for section in self.configuration.sections():
      if section == 'secret':
        continue
      for (key, val) in self.configuration.items(section):
        configs.append(section + '.' + key + '=' + val)
    return '\n'.join(configs)

  def writeConfigString(self, configString):
    configLines = configString.split('\n')
    newConfiguration = configparser.ConfigParser()
    newConfiguration.add_section('secret')
    for (key, val) in self.configuration.items('secret'):
      newConfiguration.set('secret', key, val.replace('%', '%%'))
    for configLine in configLines:
      splitted = configLine.split('.', 1)
      section = splitted[0]
      if section == 'secret':
        continue
      splitted = splitted[1].split('=', 1)
      key = splitted[0]
      value = splitted[1]
      if not newConfiguration.has_section(section):
        newConfiguration.add_section(section)
      newConfiguration.set(section, key, value)
    with open(self.configPath, 'w') as cfgfile:
      newConfiguration.write(cfgfile)
    self.configuration = newConfiguration




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""  
Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors

SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import json
import logging
import time

from c8ydm.core.apt_package_manager import AptPackageManager
from c8ydm.framework.modulebase import Initializer, Listener
from c8ydm.framework.smartrest import SmartRESTMessage


class SoftwareManager(Listener, Initializer):
    logger = logging.getLogger(__name__)
    apt_package_manager = AptPackageManager()

    def group(self, seq, sep):
        result = [[]]
        for e in seq:
            #logging.info("e: "+str(e) +" sep: " + str(sep))
            if sep not in str(e):
                result[-1].append(e)
            else:
                result[-1].append(e[:e.find(sep)])
                result.append([])

        if result[-1] == []:
            result.pop()  # thx iBug's comment
        return result

    def handleOperation(self, message):
        try:
            if 's/ds' in message.topic and message.messageId == '528':
                # When multiple operations received just take the first one for further processing
                #self.logger.debug("message received :" + str(message.values))
                messages = self.group(message.values, '\n')[0]
                deviceId = messages.pop(0)
                self.logger.info('Software update for device ' +
                                 deviceId + ' with message ' + str(messages))
                executing = SmartRESTMessage(
                    's/us', '501', ['c8y_SoftwareUpdate'])
                self.agent.publishMessage(executing)
                softwareToInstall = [messages[x:x + 4]
                                     for x in range(0, len(messages), 4)]
                errors = self.apt_package_manager.install_software(
                    softwareToInstall, True)
                self.logger.info('Finished all software update')
                if len(errors) == 0:
                    # finished without errors
                    finished = SmartRESTMessage(
                        's/us', '503', ['c8y_SoftwareUpdate'])
                else:
                    # finished with errors
                    finished = SmartRESTMessage(
                        's/us', '502', ['c8y_SoftwareUpdate', ' - '.join(errors)])
                self.agent.publishMessage(finished)
                self.agent.publishMessage(
                    self.apt_package_manager.getInstalledSoftware(False))

            if 's/ds' in message.topic and message.messageId == '516':
                # When multiple operations received just take the first one for further processing
                #self.logger.debug("message received :" + str(message.values))
                messages = self.group(message.values, '\n')[0]
                #self.logger.info("message processed:" + str(messages))
                deviceId = messages.pop(0)
                self.logger.info('Software update for device ' +
                                 deviceId + ' with message ' + str(messages))
                executing = SmartRESTMessage(
                    's/us', '501', ['c8y_SoftwareList'])
                self.agent.publishMessage(executing)
                softwareToInstall = [messages[x:x + 3]
                                     for x in range(0, len(messages), 3)]
                errors = self.apt_package_manager.installSoftware(
                    softwareToInstall, True)
                self.logger.info('Finished all software update')
                if len(errors) == 0:
                    # finished without errors
                    finished = SmartRESTMessage(
                        's/us', '503', ['c8y_SoftwareList'])
                else:
                    # finished with errors
                    finished = SmartRESTMessage(
                        's/us', '502', ['c8y_SoftwareList', ' - '.join(errors)])
                installed_software = self.apt_package_manager.get_installed_software_json(False)
                mo_id = self.agent.rest_client.get_internal_id(self.agent.serial)
                self.agent.rest_client.update_managed_object(mo_id, json.dumps(installed_software))
                self.agent.publishMessage(finished)
                #self.agent.publishMessage(
                #    self.apt_package_manager.getInstalledSoftware(False))
        except Exception as e:
            self.logger.exception(e)
            failed = SmartRESTMessage(
                's/us', '502', ['c8y_SoftwareList', str(e)])
            self.agent.publishMessage(failed)
            failed = SmartRESTMessage(
                's/us', '502', ['c8y_SoftwareUpdate', str(e)])
            self.agent.publishMessage(failed)

    def getSupportedOperations(self):
        return ['c8y_SoftwareUpdate', 'c8y_SoftwareList']

    def getSupportedTemplates(self):
        return []

    def getMessages(self):
        installed_software = self.apt_package_manager.get_installed_software_json(False)
        mo_id = self.agent.rest_client.get_internal_id(self.agent.serial)
        self.agent.rest_client.update_managed_object(mo_id, json.dumps(installed_software))
        return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""  
Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors

SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import logging, time, json
from c8ydm.framework.modulebase import Initializer
from c8ydm.framework.smartrest import SmartRESTMessage
from os.path import expanduser
import pathlib

class SmartRestInitializer(Initializer):
    logger = logging.getLogger(__name__)

    def getMessages(self):
        try:
            template_id = None
            self.logger.info(f'SmartRest Template Initializer called...')
            home = expanduser('~')
            path = pathlib.Path(home + '/.cumulocity')
            smart_rest_template_path = pathlib.Path(path / 'DM_Agent.json')
            if not smart_rest_template_path.is_file():
                self.logger.warn(f'DM_Agent.json not found in path {path}. Could not upload SmartREST Template.')
                return []
            with open(smart_rest_template_path) as f:
                payload = f.read()
                template_id = json.loads(payload)["__externalId"]
                self.logger.debug(f'SmartRest Template readed from file: {payload}')
            if not self.agent.rest_client.check_SmartRest_template_exists(template_id):
                self.logger.info(f'SmartRest Template does not exist, creating....')
               
                self.agent.rest_client.create_SmartRest_template(payload,template_id)
                msg = SmartRESTMessage('s/us', '400', ['c8y_SmartRestTemplateUpload', 'C8Y DM Agent Uploaded a SmartRest Template'])
                return [msg]
            else:
                self.logger.info(f'SmartRest Template found, skipping creation...')
        except Exception as ex:
            self.logger.exception(f'Error on reading and uploading SmartRest Template:', ex)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors

SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import re
from typing import Tuple
from subprocess import Popen, PIPE, TimeoutExpired

class CommandAlias:
    """Command alias class to create an alias for existing command
    """
    def __init__(self,
                 alias: str,
                 command: str,
                 usage: str = '',
                 use_stdout: bool = True,
                 use_stderr: bool = True):
        self._alias = alias
        self._command = command
        self._usage = usage
        self._use_stdout = use_stdout
        self._use_stderr = use_stderr

    def show_usage(self) -> str:
        """Show the command usage

        Returns:
            str: Usage string
        """
        if self._usage:
            return self._usage
        return self._alias

    def transform_command(self, alias: str) -> str:
        """Transform the command alias to the actual command that will be executed

        Args:
            alias (str): Command alias

        Returns:
            str: Command to be executed
        """
        if alias:
            return self._command
        return ''

    def is_match(self, alias: str) -> bool:
        """Check if the given alias matches the current command

        Args:
            alias (str): [description]

        Returns:
            bool: [description]
        """
        return self._alias == alias

    def format_output(self, stdout: str, stderr: str) -> str:
        """Format the stdout and stderr to a single string

        Args:
            stdout (str): Standard output
            stderr (str): Standard error

        Returns:
            str: output text
        """
        output_text = ''

        if stdout and self._use_stdout:
            output_text += stdout

        if stderr and self._use_stderr:
            output_text += stderr
        return output_text


    def execute_command(self, alias: str, timeout: int = 10) -> Tuple[int, str]:
        """Execute a command

        Args:
            alias (str): Command alias which will be transformed to the real command
            timeout (int, optional): Timeout in seconds. Defaults to 10.

        Raises:
            CommandFailedError: Command failed to execute error (non-zero exit code)
            CommandTimeoutError: Command timed out error

        Returns:
            (int, str): Command output in the format of (exit_code, output)
        """
        command = self.transform_command(alias)

        full_command = ['/bin/bash', '-c', f'{command}']
        process = Popen(full_command, shell=False, stdin=PIPE, stdout=PIPE, stderr=PIPE, close_fds=True)

        try:
            stdout, stderr = process.communicate(timeout=timeout)
            exit_code = process.returncode

            output_text = self.format_output(stdout.decode('utf8'), stderr.decode('utf8'))

            if exit_code != 0:
                raise CommandFailedError('command: {self._alias}, exit_code={exit_code},')

            return (exit_code, output_text)
        except TimeoutExpired:
            raise CommandTimeoutError(f'command: {self._alias}')


class CommandAliasWithArgs(CommandAlias):
    """Create a command alias that uses arguments

    Example:
    > CommandAliasWithArgs('show logs (\\w+)', r'journal -u "\\1" -n 100')

    Args:
        CommandAlias ([type]): [description]
    """
    def __init__(self, alias: str, command: str, usage: str = ''):
        super().__init__(alias, command, usage)
        self._template_pattern = re.compile(alias)

    def transform_command(self, alias) -> str:
        return re.sub(self._alias, self._command, alias)

    def is_match(self, alias) -> bool:
        return self._template_pattern.match(alias) is not None

class InvalidCommandError(Exception):
    """Command not found error"""

    details: str = 'Invalid pre-defined command'

    def __init__(self, *args, **kwargs):
        failure_reason: str = (
            f'Invalid command. Command execution is limited to set of pre-defined commands'
        )
        super().__init__(
            failure_reason=failure_reason,
            details=self.details,
            *args,
            **kwargs
        )


class CommandFailedError(Exception):
    """Command failed error"""

    details: str = 'Command failed to execute successfully'

    def __init__(self, *args, **kwargs):
        failure_reason: str = (
            f'A non-zero code was returned by the command.'
        )
        super().__init__(
            failure_reason=failure_reason,
            details=self.details,
            *args,
            **kwargs
        )

class CommandTimeoutError(Exception):
    """Command timeout error"""

    details: str = 'Command timed out before it was finished'

    def __init__(self, *args, **kwargs):
        failure_reason: str = (
            f'Command timeout. ' \
            f'Command are only permitted to run for a short amount of time as defined in the configuration'
        )
        super().__init__(
            failure_reason=failure_reason,
            details=self.details,
            *args,
            **kwargs
        )



"""  
Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors

SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import logging
from c8ydm.framework.smartrest import SmartRESTMessage
from c8ydm.framework.modulebase import Listener
from c8ydm.core.apt_package_manager import AptPackageManager

class DeviceProfileListener(Listener):
    logger = logging.getLogger(__name__)
    fragment = 'c8y_DeviceProfile'
    device_profiles_message_id = '527'
    apt_package_manager = AptPackageManager()

    def _set_executing(self):
        executing = SmartRESTMessage('s/us', '501', [self.fragment])
        self.agent.publishMessage(executing)

    def _set_success(self):
        success = SmartRESTMessage('s/us', '503', [self.fragment])
        self.agent.publishMessage(success)

    def _set_failed(self, reason):
        failed = SmartRESTMessage('s/us', '502', [self.fragment, reason])
        self.agent.publishMessage(failed)
    
    def _apply_device_profile(self, id):
        msg = None
        if id == None:
            msg = SmartRESTMessage('s/us', '121', ['true'])
        else:
            msg = SmartRESTMessage('s/us', '121', ['true', id])
        self.agent.publishMessage(msg)
    
    def _install_software_packages(self, messages):
        softwareToInstall = [messages[x:x + 4]
                                     for x in range(0, len(messages), 4)]
        self.logger.info(f'Software will be changed: {softwareToInstall}')
        errors = self.apt_package_manager.install_software(
                    softwareToInstall, True)
        return errors
    
    def _process_device_profile_msg(self, message):
        operation_values = {'$FW': [], '$SW': [], '$CONF': []}
        active_operation = ''

        for value in message.values:
            active_operation = '$FW' if value == '$FW' else active_operation
            active_operation = '$SW' if value == '$SW' else active_operation
            active_operation = '$CONF' if value == '$CONF' else active_operation

            if active_operation and not value.startswith('$'):
                operation_values[active_operation].append(value)
                    
        
        # TODO Firmware Block 
               
        # Software Block
        if len(operation_values['$SW']) > 0:
            self.logger.debug(f'Software Operations of Device Profile: {operation_values["$SW"]}')
            errors = self._install_software_packages(operation_values['$SW'])
            if len(errors) == 0:
                self._set_success()
            else:
                # finished with errors
                self._set_failed(errors)
            self.agent.publishMessage(
                self.apt_package_manager.getInstalledSoftware(True))
                    
        #TODO Configuration Block

    
    def handleOperation(self, message):
        """Callback that is executed for any operation received

            Raises:
            Exception: Error when handling the operation
        """
        # Example Msg 527,1da499426b9e,$SW,nano,latest,test,install
        if 's/ds' in message.topic and self.device_profiles_message_id == message.messageId:
            try:
                messages = message.values
                self._set_executing()
                self.logger.info(f'Device Profile Message received: {messages}')
                
                self._apply_device_profile(None)
                self._process_device_profile_msg(message)
                self._set_success()
            except Exception as ex:
                self.logger.exception(ex)
                self._set_failed(str(ex))

    def getSupportedOperations(self):
        return [self.fragment]

    def getSupportedTemplates(self):
        return []





#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""  
Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors

SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import logging, time
import subprocess
from c8ydm.framework.modulebase import Sensor, Initializer, Listener
from c8ydm.framework.smartrest import SmartRESTMessage
from c8ydm.core.docker_watcher import DockerWatcher

class DockerSensor(Sensor, Initializer, Listener):
    xid = 'c8y-dm-agent-v1.0'
    logger = logging.getLogger(__name__)
    docker_watcher = DockerWatcher()
    fragment = 'c8y_Docker'

    def getSensorMessages(self):
        #self.logger.info(f'Docker Update Loop called...')
        payload = self.docker_watcher.get_stats()
        internal_id = self.agent.rest_client.get_internal_id(self.agent.serial)
        self.agent.rest_client.update_managed_object(internal_id, payload)
        return []

    def getMessages(self):
        self.logger.info(f'Docker Initializer called...')
        payload = self.docker_watcher.get_stats()
        internal_id = self.agent.rest_client.get_internal_id(self.agent.serial)
        self.agent.rest_client.update_managed_object(internal_id, payload)
        return []
    
    def _set_executing(self):
        executing = SmartRESTMessage('s/us', '501', [self.fragment])
        self.agent.publishMessage(executing)

    def _set_success(self):
        success = SmartRESTMessage('s/us', '503', [self.fragment])
        self.agent.publishMessage(success)

    def _set_failed(self, reason):
        failed = SmartRESTMessage('s/us', '502', [self.fragment, reason])
        self.agent.publishMessage(failed)
    
    def handleOperation(self, message):
        if message.messageId == 'dm501':
            self.logger.info('Found a c8y_Docker operation')
            try:
                self._set_executing()
                command = message.values[1]
                create_name = message.values[2]
                image = message.values[3]
                ports = message.values[4]
                container_id = message.values[5]
                update_name = message.values[6]
                if command == 'create':
                    process = subprocess.Popen(["docker","run","-d","--name",create_name,"-p",ports,image],stdout=subprocess.PIPE,stderr=subprocess.PIPE)
                elif command == 'delete':
                    process = subprocess.Popen(["docker","rm",update_name],stdout=subprocess.PIPE,stderr=subprocess.PIPE)
                elif command == 'restart':
                    process = subprocess.Popen(["docker","restart",update_name],stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                elif command == 'stop':
                    process = subprocess.Popen(["docker","stop",update_name],stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                elif command == 'start':
                    process = subprocess.Popen(["docker","start",update_name],stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                process.wait()
                if process.returncode == 0:
                    self._set_success()
                    payload = self.docker_watcher.get_stats()
                    internal_id = self.agent.rest_client.get_internal_id(self.agent.serial)
                    self.agent.rest_client.update_managed_object(internal_id, payload)
                else:
                    stderr = str(process.stderr.read().decode('utf-8'))
                    self.logger.error(f'Following error raised for docker: {stderr}')
                    self._set_failed(stderr)
            except Exception as e:
                self.logger.error(f'The following error occured:{e}')
                self._set_failed(stderr)
    
    def getSupportedOperations(self):
        return [self.fragment]

    def getSupportedTemplates(self):
        return [self.xid]


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""  
Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors

SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import logging, time, uuid
import socket
import ipaddress
import requests
from c8ydm.framework.modulebase import Initializer
from c8ydm.framework.smartrest import SmartRESTMessage

class Network(Initializer):
    """ Network Module"""
    logger = logging.getLogger(__name__)
    net_message_id = 'dm100'
    pos_message_id = '402'
    xid = 'c8y-dm-agent-v1.0'
    def_adapter = None

    def getMessages(self):
        net_msg = None
        pos_msg = None
        try:
            self.logger.info(f'Network Initializer called...')
            ip = socket.gethostbyname(socket.gethostname())
            netmask = ipaddress.IPv4Network(ip).netmask
            name = None
            mac = self.get_mac()
            for interface in socket.if_nameindex():
                if_name = interface[1]
                if  if_name.startswith('eth'):
                    name = interface[1]
                    break
            enabled = 1
            net_msg = SmartRESTMessage('s/uc/'+self.xid, self.net_message_id, [self.serial, ip, netmask, name, enabled, mac])
            geo_data = self.get_geo_data()
            if geo_data and geo_data['latitude'] is not None and geo_data['longitude'] is not None:
                pos_msg = SmartRESTMessage('s/us', self.pos_message_id, [geo_data['latitude'], geo_data['longitude']])
            return [net_msg, pos_msg]
        except Exception as ex:
            self.logger.error(f'Error on retrieving Network Details: {ex}')
            return None
        
    def get_geo_data(self):
        """ Retrieves GEO-Data """
        try:
            geo_data = requests.get(f'https://ipapi.co/json/').json()
            self.logger.debug(f'Geo-Data: {geo_data}')
            return geo_data
        except Exception as ex:
            self.logger.error(f'Error retrieving Geodata: {ex}')
            return None
    
    def get_mac(self):
        """ Get Mac Address"""
        mac_num = hex(uuid.getnode()).replace('0x', '').replace('L', '').upper()
        mac_num = mac_num.zfill(12)
        mac = ':'.join(mac_num[i: i + 2] for i in range(0, 11, 2))
        return mac





#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""  
Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors

SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import json
import logging
import time

from c8ydm.core.firmware_manager import FirmwareManager
from c8ydm.framework.modulebase import Initializer, Listener
from c8ydm.framework.smartrest import SmartRESTMessage


class FirmwareManagement(Listener, Initializer):
    logger = logging.getLogger(__name__)
    fragment = 'c8y_Firmware'
    firmware_manager = FirmwareManager()

    def group(self, seq, sep):
        result = [[]]
        for e in seq:
            if sep not in str(e):
                result[-1].append(e)
            else:
                result[-1].append(e[:e.find(sep)])
                result.append([])
        if result[-1] == []:
            result.pop()
        return result
    
    def _set_executing(self):
        executing = SmartRESTMessage('s/us', '501', [self.fragment])
        self.agent.publishMessage(executing)

    def _set_success(self):
        success = SmartRESTMessage('s/us', '503', [self.fragment])
        self.agent.publishMessage(success)

    def _set_failed(self, reason):
        failed = SmartRESTMessage('s/us', '502', [self.fragment, reason])
        self.agent.publishMessage(failed)

    def handleOperation(self, message):
        try:
            is_simulated = self.agent.simulated
            if 's/ds' in message.topic and message.messageId == '515':
                messages = self.group(message.values, '\n')[0]
                deviceId = messages.pop(0)
                self.logger.info('Firmware Update for device ' +
                                 deviceId + ' with message ' + str(messages))
                self._set_executing()
                firmwareToInstall = [messages[x:x + 3]
                                     for x in range(0, len(messages), 3)]
                #TODO Handle Firmware Update properly. Not working for Docker and Raspberry Pi.
                self._set_failed('Firmware Update triggered but not supported by this device type')
                alarmMsg = SmartRESTMessage('s/us', '304', ['c8y_FirmwareUpdateAlarm','Firmware Update triggered but not supported by this device type',''])
                #eventMsg = SmartRESTMessage('s/us', '400', ['c8y_FirmwareUpdateEvent', 'Firmware could not be updated as it is not supported by the current Device.'])
                self.agent.publishMessage(alarmMsg)
                """ if is_simulated:
                    #No Firmware Updates are currently supported for docker containers
                    self._set_failed('No Firmware Updates are currently supported for docker containers!')
                else:
                    #TODO Update Firmware for specific devices!
                    self.logger.info('Finished firmware update')
                    #TODO Update Firmware Version
                """
        except Exception as e:
            self._set_failed(e)

    def getSupportedOperations(self):
        return ['c8y_Firmware']

    def getSupportedTemplates(self):
        return []

    def getMessages(self):
        #TODO Check current Firmware version, Update Operation, Update Fragment
        return self.get_firmware_msg()
    
    def get_firmware_msg(self):
        current_kernel_version = self.firmware_manager.get_current_kernel_version()
        dist_name = self.firmware_manager.get_dist_name()
        dist_version = self.firmware_manager.get_dist_version()
        firmware_name = dist_name + "-" + dist_version
        firmware_version = current_kernel_version
        firmware_msg = SmartRESTMessage('s/us', '115', [firmware_name, firmware_version, ''])
        return [firmware_msg]



import logging, time
import shlex
from c8ydm.framework.modulebase import Sensor, Initializer, Listener
from c8ydm.framework.smartrest import SmartRESTMessage
try:
    from sense_hat import SenseHat, ACTION_PRESSED
    SENSE = SenseHat()
except Exception as e:
    print('SenseHat not available')
    SENSE = None

class DeviceSensor(Sensor, Initializer, Listener):
    """ SenseHAT Module """
    logger = logging.getLogger(__name__)
    fragment = 'c8y_Message'
    message_id = 'dm502'
    xid = 'c8y-dm-agent-v1.0'

    def _set_executing(self):
        executing = SmartRESTMessage('s/us', '501', [self.fragment])
        self.agent.publishMessage(executing)

    def _set_success(self):
        success = SmartRESTMessage('s/us', '503', [self.fragment])
        self.agent.publishMessage(success)

    def _set_failed(self, reason):
        failed = SmartRESTMessage('s/us', '502', [self.fragment, reason])
        self.agent.publishMessage(failed)

    def handleOperation(self, message):
        """Callback that is executed for any operation received

            Raises:
            Exception: Error when handling the operation
       """
        try:
            #self.logger.debug(
            #    f'Handling Cloud Remote Access operation: listener={__name__}, message={message}')

            if message.messageId == self.message_id:
                self._set_executing()
                message = message.values[1]
                self.display_message(message)
                self._set_success()
    
        except Exception as ex:
            self.logger.error(f'Handling operation error. exception={ex}')
            self._set_failed(str(ex))
            # raise


    def getSensorMessages(self):
        try:
            return self.send_stats()
        except Exception as e:
            self.logger.exception(f'Error in SenseHAT getSensorMessages: {e}', e)

    def getMessages(self):
        try:
            self.listenForJoystick()
            return self.send_stats()
        except Exception as e:
            self.logger.exception(f'Error in SenseHAT getMessages: {e}', e)
    
    def send_stats(self):
        if SENSE:
            self.stats = []
            self.stats.append(SmartRESTMessage('s/us', '200', ['SenseHat', 'Temperature', SENSE.get_temperature()]))
            self.stats.append(SmartRESTMessage('s/us', '200', ['SenseHat', 'Humidity', SENSE.get_humidity()]))
            self.stats.append(SmartRESTMessage('s/us', '200', ['SenseHat', 'Pressure', SENSE.get_pressure()]))
            acceleration = SENSE.get_accelerometer_raw()
            acc_x = acceleration['x']
            acc_y = acceleration['y']
            acc_z = acceleration['z']
            self.stats.append(SmartRESTMessage('s/us', '200', ['SenseHat', 'Acceleration.xValue', acc_x]))
            self.stats.append(SmartRESTMessage('s/us', '200', ['SenseHat', 'Acceleration.yValue', acc_y]))
            self.stats.append(SmartRESTMessage('s/us', '200', ['SenseHat', 'Acceleration.zValue', acc_z]))
            gyroscope = SENSE.gyro_raw
            gyro_x = gyroscope["x"]
            gyro_y = gyroscope["y"]
            gyro_z = gyroscope["z"]
            self.stats.append(SmartRESTMessage('s/us', '200', ['SenseHat', 'Gyroscope.xValue', gyro_x]))
            self.stats.append(SmartRESTMessage('s/us', '200', ['SenseHat', 'Gyroscope.yValue', gyro_y]))
            self.stats.append(SmartRESTMessage('s/us', '200', ['SenseHat', 'Gyroscope.zValue', gyro_z]))
            compass = SENSE.compass_raw
            compass_x = compass["x"]
            compass_y = compass["y"]
            compass_z = compass["z"]
            self.stats.append(SmartRESTMessage('s/us', '200', ['SenseHat', 'Compass.xValue', compass_x]))
            self.stats.append(SmartRESTMessage('s/us', '200', ['SenseHat', 'Compass.yValue', compass_y]))
            self.stats.append(SmartRESTMessage('s/us', '200', ['SenseHat', 'Compass.zValue', compass_z]))
            return self.stats
        else:
            return []
    

    def display_message(self,message):
        if SENSE:
            messageArray =  shlex.shlex(message, posix=True)
            messageArray.whitespace =',' 
            messageArray.whitespace_split =True 
            msg = str(list(messageArray)[-1])
            self.logger.info(f'Display message: {msg}')
            SENSE.show_message(msg)
            SENSE.clear
    
    def joystick_up(self, event):
        if event.action == ACTION_PRESSED:
            msg = SmartRESTMessage('s/us', '400', ['c8y_SenseHATJoystickUpEvent', 'Joystick Up Event!'])
            self.agent.publishMessage(msg)
    
    def joystick_down(self, event):
        if event.action == ACTION_PRESSED:
            msg = SmartRESTMessage('s/us', '400', ['c8y_SenseHATJoystickDownEvent', 'Joystick Down Event!'])
            self.agent.publishMessage(msg)

    def joystick_left(self, event):
        if event.action == ACTION_PRESSED:
            msg = SmartRESTMessage('s/us', '400', ['c8y_SenseHATJoystickLeftEvent', 'Joystick Left Event!'])
            self.agent.publishMessage(msg)

    def joystick_right(self, event):
        if event.action == ACTION_PRESSED:
            msg = SmartRESTMessage('s/us', '400', ['c8y_SenseHATJoystickRightEvent', 'Joystick Right Event!'])
            self.agent.publishMessage(msg)

    def joystick_middle(self, event):
        if event.action == ACTION_PRESSED:
            msg = SmartRESTMessage('s/us', '400', ['c8y_SenseHATJoystickMiddleEvent', 'Joystick Middle Event!'])
            self.agent.publishMessage(msg)

    def listenForJoystick(self):
        if SENSE:
            SENSE.stick.direction_up = self.joystick_up
            SENSE.stick.direction_down = self.joystick_down
            SENSE.stick.direction_left = self.joystick_left
            SENSE.stick.direction_right = self.joystick_right
            SENSE.stick.direction_middle = self.joystick_middle
    
    def getSupportedOperations(self):
        return [self.fragment]
    
    def getSupportedTemplates(self):
        return [self.xid]


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors

SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import logging
import os
import platform
import distro

class FirmwareManager:
    logger = logging.getLogger(__name__)

    def get_current_kernel_version(self):
        if 'Windows' == platform.system():
            return platform.uname().version
        else:
            os_info = platform.uname()
            kernel = os_info.release
            return kernel
    
    def get_dist_name(self):
        if 'Windows' == platform.system():
            return platform.system()
        else:
            return self._get_linux_dist_name()

    def get_dist_version(self):
        if 'Windows' == platform.system():
            return platform.uname().release
        else:
            return self._get_linux_dist_version()

    def _get_linux_dist_name(self):
        return distro.id()
    
    def _get_linux_dist_version(self):
        return distro.version()


#
# Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from setuptools import setup, find_packages
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
      python_requires=">=3.7",
      name='c8ydm',
      long_description=long_description,
      long_description_content_type="text/markdown",
      version='1.1.0',
      description='Cumulocity Device Management Agent',
      author='Tobias Sommer, Stefan Witschel, Marco Stoffel, Murat Bayram',
      author_email="Stefan.Witschel@softwareag.com",
      license='Apache 2.0',
      packages=find_packages(),
      entry_points={
        'console_scripts': [
              'c8ydm.start=c8ydm.main:start',
              'c8ydm.stop=c8ydm.main:stop'
            ],
      },
      install_requires=[
        "paho-mqtt>=1.5.1",
        "psutil>=5.8.0",
        "requests>=2.25.1",
        "c8y-device-proxy>=1.1.1",
        "certifi>=2021.10.8",
        "distro>=1.6.0",
        "sense-hat>=2.2.0"
    ],
      zip_safe=False)


"""  
Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors

SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import logging
import time
import ssl
import _thread
import threading
import certifi
import paho.mqtt.client as mqtt


import c8ydm.utils.moduleloader as moduleloader
from c8ydm.client.rest_client import RestClient
from c8ydm.core.configuration import ConfigurationManager
from c8ydm.framework.smartrest import SmartRESTMessage


class Agent():
    __sensors = []
    __listeners = []
    __supportedOperations = set()
    __supportedTemplates = set()
    stopmarker = 0

    def __init__(self, serial, path, configuration, pidfile, simulated):
        self.logger = logging.getLogger(__name__)
        self.serial = serial
        self.simulated = simulated
        self.__client = mqtt.Client(serial)
        self.configuration = configuration
        self.pidfile = pidfile
        self.path = path
        self.url = self.configuration.getValue('mqtt', 'url')
        self.port = self.configuration.getValue('mqtt', 'port')
        self.ping = self.configuration.getValue(
            'mqtt', 'ping.interval.seconds')
        self.tls = self.configuration.getBooleanValue('mqtt', 'tls')
        #self.cacert = self.configuration.getValue('mqtt', 'cacert')
        self.cert_auth = self.configuration.getBooleanValue(
            'mqtt', 'cert_auth')
        self.client_cert = self.configuration.getValue('mqtt', 'client_cert')
        self.client_key = self.configuration.getValue('mqtt', 'client_key')
        self.interval = int(self.configuration.getValue(
            'agent', 'main.loop.interval.seconds'))
        self.device_name = f'{self.configuration.getValue("agent", "name")}-{serial}'
        self.device_type = self.configuration.getValue('agent', 'type')

        self.stop_event = threading.Event()
        self.refresh_token_interval = 60
        self.token = None
        self.is_connected = False
        self.rest_client = RestClient(self)

        if self.simulated:
            self.model = 'docker'
        else:
            self.model = 'raspberry'

    def handle_sensor_message(self, sensor):
        messages = sensor.getSensorMessages()
        if messages is not None and len(messages) > 0:
            for message in messages:
                self.publishMessage(message)

    def handle_initializer_message(self, initializer):
        messages = initializer.getMessages()
        if messages is not None and len(messages) > 0:
            for message in messages:
                if message:
                    self.logger.debug('Send topic: %s, msg: %s',
                                    message.topic, message.getMessage())
                    self.publishMessage(message)

    def run(self):
        try:
            self.logger.info('Starting agent')
            self.__client = mqtt.Client(self.serial)
            credentials = self.configuration.getCredentials()
            self.__client = self.connect(
                credentials, self.serial, self.url, int(self.port), int(self.ping))
            self.__client.loop_start()
            while not self.is_connected:
                time.sleep(1)
                self.logger.debug('Waiting for MQTT Client to be connected')
            self.__init_agent()
            while not self.stopmarker:
                self.logger.debug('New cycle')
                self.interval = int(self.configuration.getValue(
                    'agent', 'main.loop.interval.seconds'))
                for sensor in self.__sensors:
                    sensor_thread = threading.Thread(target=self.handle_sensor_message, args=(sensor,))
                    sensor_thread.daemon = True
                    sensor_thread.name = f'SensorThread-{sensor.__class__.__name__}'
                    sensor_thread.start()
                    #_thread.start_new_thread(self.handle_sensor_message, (sensor,))
                time.sleep(self.interval)
        except Exception as e:
            self.logger.exception(f'Error in C8Y Agent: {e}', e)
            self.disconnect(self.__client)
            self.logger.info('Will retry to connect to C8Y in 5 sec...')
            time.sleep(5)
            # Run again after 5 sec. delay.
            self.run()

    def connect(self, credentials, serial, url, port, ping):
        try:
            self.__client.on_connect = self.__on_connect
            self.__client.on_message = self.__on_message
            self.__client.on_disconnect = self.__on_disconnect
            #self.__client.on_subscribe = self.__on_subscribe
            self.__client.on_log = self.__on_log

            if self.tls:
                if self.cert_auth:
                    self.logger.debug('Using certificate authenticaiton')
                    self.__client.tls_set(certifi.where(),
                                          certfile=self.client_cert,
                                          keyfile=self.client_key,
                                          tls_version=ssl.PROTOCOL_TLSv1_2,
                                          cert_reqs=ssl.CERT_NONE
                                          )
                else:
                    self.__client.tls_set(certifi.where())
                    self.__client.username_pw_set(
                        credentials[0]+'/' + credentials[1], credentials[2])
            else:
                self.__client.username_pw_set(
                    credentials[0]+'/' + credentials[1], credentials[2])

            self.__client.connect(url, int(port), int(ping))
            self.__client.loop_start()
            return self.__client
        except Exception as e:
            self.logger.exception(f'Error on connecting C8Y Agent: {e}', e)
            self.disconnect(self.__client)
            self.logger.info('Will retry to connect to C8Y in 5 sec...')
            time.sleep(5)
            # Run again after 5 sec. delay.
            return self.run()

    def disconnect(self, client):
        self.logger.info("Disconnecting MQTT Client")
        self.__client = None
        if client == None:
            return
        client.loop_stop()  # stop the loop
        client.disconnect()
        if self.cert_auth:
            self.logger.info("Stopping refresh token thread")
            self.stop_event.set()
        

    def stop(self):
        msg = SmartRESTMessage('s/us', '400', ['c8y_AgentStopEvent', 'C8Y DM Agent stopped'])
        self.publishMessage(msg, qos=0, wait_for_publish=True)
        self.disconnect(self.__client)
        self.stopmarker = 1

    def pollPendingOperations(self):
        while not self.stopmarker:
            try:
                time.sleep(15)
                self.logger.debug('Polling for pending Operations')
                pending = SmartRESTMessage('s/us', '500', [])
                self.publishMessage(pending)
            except Exception as e:
                self.logger.error(
                    'Error on polling for Pending Operations: ' + str(e))

    def __init_agent(self):
        self.__listeners = []
        self.__sensors = []
        # set Device Name
        msg = SmartRESTMessage('s/us', '100', [self.device_name, self.device_type])
        self.publishMessage(msg, 2, wait_for_publish=True)
        #self.__client.publish(
        #    "s/us", "100,"+self.device_name+","+self.device_type, 2).wait_for_publish()
        #self.logger.info(f'Device published!')
        configurationManager = ConfigurationManager(
            self.serial, self, self.configuration)

        messages = configurationManager.getMessages()
        for message in messages:
            self.logger.debug('Send topic: %s, msg: %s',
                              message.topic, message.getMessage())
            self.__client.publish(message.topic, message.getMessage())
        self.__listeners.append(configurationManager)
        self.__supportedOperations.update(
            configurationManager.getSupportedOperations())

        # Load custom modules
        modules = moduleloader.findAgentModules()
        classCache = {}

        for sensor in modules['sensors']:
            currentSensor = sensor(self.serial, self)
            classCache[sensor.__name__] = currentSensor
            self.__sensors.append(currentSensor)
        for listener in modules['listeners']:
            if listener.__name__ in classCache:
                currentListener = classCache[listener.__name__]
            else:
                currentListener = listener(self.serial, self)
                classCache[listener.__name__] = currentListener
            supportedOperations = currentListener.getSupportedOperations()
            supportedTemplates = currentListener.getSupportedTemplates()
            if supportedOperations is not None:
                self.__supportedOperations.update(supportedOperations)
            if supportedTemplates is not None:
                self.__supportedTemplates.update(supportedTemplates)
            self.__listeners.append(currentListener)
        for initializer in modules['initializers']:
            if initializer.__name__ in classCache:
                currentInitializer = classCache[initializer.__name__]
            else:
                currentInitializer = initializer(self.serial, self)
                classCache[initializer.__name__] = currentInitializer
            init_thread = threading.Thread(target=self.handle_initializer_message, args=(currentInitializer,))
            init_thread.daemon = True
            init_thread.name = f'InitializerThread-{currentInitializer.__class__.__name__}'
            init_thread.start()
            #_thread.start_new_thread(self.handle_initializer_message, (currentInitializer,))

        classCache = None

        # set supported operations
        self.logger.info('Supported operations:')
        self.logger.info(self.__supportedOperations)
        supportedOperationsMsg = SmartRESTMessage(
            's/us', 114, self.__supportedOperations)
        self.publishMessage(supportedOperationsMsg)

        # set required interval
        required_interval = self.configuration.getValue(
            'agent', 'requiredinterval')
        self.logger.info(f'Required interval: {required_interval}')
        requiredIntervalMsg = SmartRESTMessage(
            's/us', 117, [f'{required_interval}'])
        self.publishMessage(requiredIntervalMsg)

        # set device model
        self.logger.info('Model:')
        self.logger.info([self.serial, self.model, '1.0'])
        modelMsg = SmartRESTMessage(
            's/us', 110, [self.serial, self.model, '1.0'])
        self.publishMessage(modelMsg)

        self.__client.subscribe('s/e')
        self.__client.subscribe('s/ds')
        self.__client.subscribe('s/dat')

        # subscribe additional topics
        for xid in self.__supportedTemplates:
            self.logger.info('Subscribing to XID: %s', xid)
            self.__client.subscribe('s/dc/' + xid)
        if self.cert_auth:
            self.logger.info("Starting refresh token thread ")
            token_thread = threading.Thread(target=self.refresh_token)
            token_thread.daemon = True
            token_thread.name = f'TokenThread-1'
            token_thread.start()
            #_thread.start_new_thread(self.refresh_token)
            # refresh_token_thread.start()
        # Set all dangling Operations to failed on Agent start
        internald_id = self.rest_client.get_internal_id(self.serial)
        ops = self.rest_client.get_all_dangling_operations(internald_id)
        self.rest_client.set_operations_to_failed(ops)


    def __on_connect(self, client, userdata, flags, rc):
        try:
            self.logger.info('Agent connected with result code: ' + str(rc))
            if rc > 0:
                self.logger.warning(
                    'Disconnecting Agent and try to re-connect manually..')
                # TODO What should be done when rc != 0? Reconnect? Abort?
                self.disconnect(self.__client)
                time.sleep(5)
                self.logger.info('Restarting Agent ..')
                self.run()
            else:
                self.is_connected = True
        except Exception as ex:
            self.logger.error(ex)

    def __on_message(self, client, userdata, msg):
        try:
            decoded = msg.payload.decode('utf-8')
            messageParts = decoded.split(',')
            message = SmartRESTMessage(
                msg.topic, messageParts[0], messageParts[1:])
            self.logger.debug('Received: topic=%s msg=%s',
                          message.topic, message.getMessage())
            if message.messageId == '71':
                self.token = message.values[0]
                self.logger.info('New JWT Token received')
            for listener in self.__listeners:
                self.logger.debug('Trigger listener ' +
                              listener.__class__.__name__)
                listener_thread = threading.Thread(target=listener.handleOperation, args=(message,))
                listener_thread.daemon = True
                listener_thread.name = f'ListenerThread-{listener.__class__.__name__}'
                listener_thread.start()
                #_thread.start_new_thread(listener.handleOperation, (message,))
        except Exception as e:
            self.logger.error(f'Error on handling MQTT Message.', e)

    def __on_disconnect(self, client, userdata, rc):
        self.logger.debug("on_disconnect rc: " + str(rc))
        # if rc==5:
        #     self.reset()
        #     return
        if rc != 0:
            self.logger.error(f'Disconnected with result code {rc}! Trying to reconnect...')
            #self.__client.reconnect()
            time.sleep(5)
            # Run again after 5 sec. delay.
            return self.run()

    def __on_log(self, client, userdata, level, buf):
        self.logger.log(level, buf)

    def publishMessage(self, message, qos=0, wait_for_publish=False):
        self.logger.debug(f'Send: topic={message.topic} msg={message.getMessage()}')
        if self.__client is not None and self.__client.is_connected:
            if wait_for_publish:
                self.__client.publish(message.topic, message.getMessage(), qos).wait_for_publish()
            else:
                self.__client.publish(message.topic, message.getMessage(), qos)


    def refresh_token(self):
        self.stop_event.clear()
        while True:
            self.logger.info("Refreshing Token")
            self.__client.publish('s/uat','',2)
            if self.stop_event.wait(timeout=self.refresh_token_interval):
                self.logger.info("Exit Refreshing Token Thread")
                break
